package auth

import (
	"crypto/hmac"
	"crypto/sha256"
	"encoding/hex"
	"encoding/json"
	"github.com/aws/aws-lambda-go/events"
	"github.com/google/uuid"
	"github.com/thoas/go-funk"
	"gitlab.bob.co.za/bob-public-utils/bobgroup-go-utils/date_utils"
	"gitlab.bob.co.za/bob-public-utils/bobgroup-go-utils/handler_utils"
	"time"
)

type SessionToken struct {
	IP          string    `json:"ip"`
	UserAgent   string    `json:"user_agent"`
	TimeCreated time.Time `json:"time_created"`
	Token       string    `json:"session_token"`
}

// SignSessionTokenWithKey signs the session token string using the secret.
func SignSessionTokenWithKey(secretKey string, sessionTokenString string) string {
	signer := hmac.New(sha256.New, []byte(secretKey))
	signer.Write([]byte(sessionTokenString))
	return hex.EncodeToString(signer.Sum(nil))
}

// GetSessionTokenString creates a unique session token string from the API request.
func GetSessionTokenString(request events.APIGatewayProxyRequest) (string, error) {
	sessionToken := SessionToken{
		IP:          request.RequestContext.Identity.SourceIP,
		UserAgent:   handler_utils.FindHeaderValue(request.Headers, "user-agent"),
		TimeCreated: date_utils.CurrentDate().Round(time.Second),
		Token:       uuid.New().String(),
	}
	sessionTokenBytes, err := json.Marshal(sessionToken)
	return string(sessionTokenBytes), err
}

// GetSignedSessionTokenString creates a unique session token string from the API request and signs it using the secret.
func GetSignedSessionTokenString(request events.APIGatewayProxyRequest, secretKey string) (string, error) {
	sessionTokenBytes, err := GetSessionTokenString(request)
	if err != nil {
		return "", err
	}

	return SignSessionTokenWithKey(secretKey, sessionTokenBytes), nil
}

// ValidateJWTWithSessionTokens attempts to validate the JWT string by signing each session token using the secret, and
// using the resulting signed session token to validate the JWT. If the JWT can be validated using a session token, the
// JsonWebToken is returned, otherwise nil is returned. If the JWT is expired, nil is returned along with the session token.
func ValidateJWTWithSessionTokens(jsonWebTokenString string, secretKey string, sessionTokens []string) (validJsonWebToken *JsonWebToken, expiredSessionToken *string) {
	// Test each session token to find one that is valid
	for _, sessionToken := range sessionTokens {
		jsonWebToken, err := ValidateJWTWithSessionToken(jsonWebTokenString, secretKey, sessionToken)
		if err == nil {
			return &jsonWebToken, nil
		} else if err.Error() == "token has expired" {
			return nil, &sessionToken
		}
	}

	return nil, nil
}

// FindAndRemoveCurrentSessionToken attempts to validate the JWT string by signing each session token using the secret,
// and using the resulting signed session token to validate the JWT. If the JWT is successfully validated with one of
// the session tokens, the session token is removed from the slice, otherwise the original session token
// slice is returned.
func FindAndRemoveCurrentSessionToken(jsonWebTokenString string, secretKey string, sessionTokens []string) (string, []string) {
	// Test each session token to find one that is valid
	for _, sessionToken := range sessionTokens {
		_, err := ValidateJWTWithSessionToken(jsonWebTokenString, secretKey, sessionToken)
		if err == nil {
			// Remove this session token from the slice
			updatedSessionTokens := funk.FilterString(sessionTokens, func(token string) bool {
				return token != sessionToken
			})
			return sessionToken, updatedSessionTokens
		}
	}

	return "", sessionTokens
}

// RemoveInvalidatedAndOldSessionTokens removes the provided invalidated session token, and checks the age of the
// other session tokens and removes the ones that are older than the provided age.
func RemoveInvalidatedAndOldSessionTokens(sessionTokens []string, invalidatedSessionToken string, age time.Duration) []string {
	ageDurationAgo := date_utils.CurrentDate().Add(-1 * age)
	validTokens := funk.FilterString(sessionTokens, func(sessionTokenString string) bool {
		// Always remove the invalidated session token
		if sessionTokenString == invalidatedSessionToken {
			return false
		}

		var sessionToken SessionToken
		err := json.Unmarshal([]byte(sessionTokenString), &sessionToken)
		if err != nil {
			// If we can't unmarshal the token then it is not valid
			return false
		}

		// Keep the token if it hasn't expired yet
		return sessionToken.TimeCreated.In(date_utils.CurrentLocation()).After(ageDurationAgo)
	})
	return validTokens
}
